from typing import Any, Dict, Optional


class AuthAPI:
    def __init__(self, api_session):
        self._api = api_session
        self._base = "/auth"

    def login(self, username: str, password: str):
        return self._api.post(f"{self._base}/login", json_body={"username": username, "password": password})

    def register(self, username: str, display_name: str, password: str, invite_code: Optional[str] = None,
                 competition_bot_key: Optional[str] = None):
        # Route /auth/register-bot for bots
        if competition_bot_key:
            body = {
                "username": username,
                "display_name": display_name,
                "password": password,
                "competition_bot_key": competition_bot_key,
            }
            return self._api.post(f"{self._base}/register-bot", json_body=body)
        
        # Otherwise, fallback to standard registration which requires an invite_code
        body = {
            "username": username,
            "display_name": display_name,
            "password": password,
        }
        if invite_code is not None:
            body["invite_code"] = invite_code
        return self._api.post(f"{self._base}/register", json_body=body)

    def register_team(self, username: str, display_name: str, password: str, team_name: str, affiliation: str,
                      first_member_name: str, first_member_email: str):
        body = {
            "username": username,
            "display_name": display_name,
            "password": password,
            "team": {
                "name": team_name,
                "affiliation": affiliation,
                "first_member": {"name": first_member_name, "email": first_member_email},
            },
        }
        return self._api.post(f"{self._base}/register-team", json_body=body)

    def change_password(self, token_headers: Dict[str, str], new_password: str):
        return self._api.post(f"{self._base}/change-password", json_body={"new_password": new_password}, headers=token_headers)

    def logout(self, token_headers: Dict[str, str]):
        return self._api.post(f"{self._base}/logout", json_body={}, headers=token_headers)



from datetime import datetime
from typing import Any, Dict, Optional


class FeedsAPI:
    def __init__(self, api_session, headers_for_username=None):
        self._api = api_session
        self._headers_for = headers_for_username

    def list(self, agent_username: Optional[str] = None) -> Dict[str, Any]:
        headers = None
        if self._headers_for is not None and agent_username:
            headers = self._headers_for(agent_username)
        r = self._api.get("/feeds/", headers=headers)
        r.raise_for_status()
        return r.json()

    def feed(self, key: str, at: Optional[datetime] = None, agent_username: Optional[str] = None) -> Dict[str, Any]:
        params = []
        if at is not None:
            params.append(f"at={at.isoformat()}")
        qs = ("?" + "&".join(params)) if params else ""
        headers = None
        if self._headers_for is not None and agent_username:
            headers = self._headers_for(agent_username)
        r = self._api.get(f"/feeds/{key}{qs}", headers=headers)
        r.raise_for_status()
        return r.json()



from typing import Any, Dict
from urllib.parse import quote_plus


class SearchAPI:
    def __init__(self, api_session):
        self._api = api_session

    def search(self, query: str) -> Dict[str, Any]:
        q = quote_plus(query)
        r = self._api.get(f"/search?query={q}")
        r.raise_for_status()
        return r.json()



from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Dict, Optional

from .util import xdg_config_home, xdg_data_home, xdg_state_home, ensure_parent


def resolve_db_path(
    config_value: Optional[str],
    default_name: str,
    base_dir: Path,
    *,
    prefer_state: bool = False,
) -> str:
    cfg_path = Path(config_value) if config_value else Path(default_name)
    if not cfg_path.is_absolute():
        cfg_path = base_dir / cfg_path
    cfg_path = cfg_path.expanduser()
    if cfg_path.exists():
        return str(cfg_path.resolve())
    xdg_base = xdg_state_home() if prefer_state else xdg_data_home()
    xdg_path = (xdg_base / "twooter" / default_name).expanduser()
    if xdg_path.exists():
        return str(xdg_path.resolve())
    ensure_parent(cfg_path)
    return str(cfg_path.resolve())


class Config:
    def __init__(self, config_path: Optional[str]) -> None:
        self.config_path = self._resolve_config_path(config_path)

        if not self.config_path or not Path(self.config_path).is_file():
            example = {
                "base_url": "https://social.legitreal.com/api",
                "personas_db": "./personas.db",
                "tokens_db": "./tokens.db",
                "teams_db": "./teams.db",
                "competition_bot_key": "botkey",
                "team_invite_code": "teaminvitecode",
            }
            raise FileNotFoundError(
                "Config not found. Create ./config.json with fields: "
                + json.dumps(example, ensure_ascii=False))

        with open(self.config_path, encoding="utf-8") as f:
            data: Dict[str, Any] = json.load(f)
        base_dir = Path(self.config_path).parent

        bu = data.get("base_url")
        if not bu:
            raise FileNotFoundError(
                "No base_url found in config. Create an XDG config at "
                f"{xdg_config_home()/ 'twooter' / 'config.json'} with " + '{"base_url":"https://..."}')
        self.base_url = bu.rstrip("/")

        self.personas_db = resolve_db_path(data.get("personas_db"), "personas.db", base_dir)
        self.tokens_db = resolve_db_path(data.get("tokens_db"), "tokens.db", base_dir, prefer_state=True)
        self.teams_db = resolve_db_path(data.get("teams_db"), "teams.db", base_dir)

        self.competition_bot_key = data.get("competition_bot_key")
        self.team_invite_code = data.get("team_invite_code")

        self.login_path = "/auth/login"
        self.register_path = "/auth/register"

        for pth in (self.personas_db, self.tokens_db, self.teams_db):
            ensure_parent(pth)

    @staticmethod
    def _resolve_config_path(explicit: Optional[str]) -> Optional[str]:
        if explicit:
            return str(Path(explicit).expanduser())
        cwd_cfg = Path.cwd() / "config.json"
        if cwd_cfg.is_file():
            return str(cwd_cfg)
        xdg_cfg = xdg_config_home() / "twooter" / "config.json"
        if xdg_cfg.is_file():
            return str(xdg_cfg)
        return None



from __future__ import annotations

from .sdk import new, Twooter
from . import sdk as sdk
from . import cli as cli

__all__ = ["new", "Twooter", "sdk", "cli"]

try:
    from importlib import metadata as _im
    __version__ = _im.version("twooter")
except Exception:
    __version__ = "0"



from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Dict, Any, List, Union

from .cli import TwooterClient, Config

@dataclass
class TwooterOptions:
    base_url: str
    personas_db: Optional[str] = None
    tokens_db: Optional[str] = None
    teams_db: Optional[str] = None
    login_path: str = "/auth/login"
    register_path: str = "/auth/register"
    competition_bot_key: Optional[str] = None
    team_invite_code: Optional[str] = None
    debug: bool = False
    config_path: Optional[str] = None

    @staticmethod
    def from_env(*, base_url: Optional[str] = None, debug: Optional[bool] = None) -> "TwooterOptions":
        dbg = bool(debug) if debug is not None else False
        cfg = Config(None)
        return TwooterOptions(
            base_url=cfg.base_url,
            personas_db=cfg.personas_db,
            tokens_db=cfg.tokens_db,
            teams_db=cfg.teams_db,
            competition_bot_key=cfg.competition_bot_key,
            team_invite_code=cfg.team_invite_code,
            debug=dbg,
            config_path=cfg.config_path,
        )

class Twooter:
    def __init__(self, opts: TwooterOptions) -> None:
        self._opts = opts
        if not opts.personas_db or not opts.tokens_db or not opts.teams_db:
            raise ValueError("personas_db, tokens_db, and teams_db must be set; TwooterOptions.from_env to load <--> config.json")
        self._client = TwooterClient(
            base_url=opts.base_url,
            personas_db=opts.personas_db,
            tokens_db=opts.tokens_db,
            teams_db=opts.teams_db,
            login_path=opts.login_path,
            register_path=opts.register_path,
            default_invite=opts.team_invite_code,
            default_bot_key=opts.competition_bot_key,
            config_path=opts.config_path,
            debug=opts.debug,
        )
        self._agent: Optional[str] = None

    # ---------- auth/session ----------

    def login(
        self,
        username: str,
        password: str,
        *,
        display_name: Optional[str] = None,
        invite_code: Optional[str] = None,
        bot_key: Optional[str] = None,
        auto_create_if_missing: bool = True,
        team_name: Optional[str] = None,
        affiliation: Optional[str] = None,
        member_name: Optional[str] = None,
        member_email: Optional[str] = None,
    ) -> Dict[str, Any]:
        info = self._client.login(
            username=username,
            password=password,
            display_name=display_name or username,
            invite_code=invite_code or self._opts.team_invite_code,
            bot_key=bot_key or self._opts.competition_bot_key,
            try_register_on_403=auto_create_if_missing,
            auto_confirm=True,
            team_name=team_name,
            affiliation=affiliation,
            member_name=member_name or display_name or username,
            member_email=member_email,
        )
        self._agent = username
        return info

    def logout(self) -> Dict[str, Any]:
        u = self._need_agent()
        headers = self._client._auth_headers(u)
        r = self._client.auth.logout(headers)
        r.raise_for_status()
        return {"data": True}

    def change_password(self, new_password: str) -> Dict[str, Any]:
        u = self._need_agent()
        headers = self._client._auth_headers(u)
        r = self._client.auth.change_password(headers, new_password)
        r.raise_for_status()
        return {"data": True}

    def use_agent(self, username: str) -> None:
        self._agent = username

    def whoami(self) -> Dict[str, Any]:
        u = self._need_agent()
        try:
            me = self._client.users.me(u)
        except Exception:
            me = {"data": None}
        try:
            team = self._client.competition.get(u)
        except Exception:
            team = {"data": None}

        data = me.get("data") if isinstance(me, dict) else None
        username = data.get("username") if isinstance(data, dict) else u
        email_verified = data.get("email_verified") if isinstance(data, dict) else None
        role = (data.get("role") if isinstance(data, dict) else None) or None
        team_data = team.get("data") if isinstance(team, dict) else None
        if role is None:
            role = "TEAM_ADMIN" if team_data else "COMPETITOR"
        return {"data": {"username": username, "role": role, "email_verified": email_verified, "team": team_data}}

    def token_info(self) -> Dict[str, Any]:
        u = self._need_agent()
        info = self._client.tokens.get_info(u) or {}
        if not info:
            return {"data": None}
        return {
            "data": {
                "username": info.get("username", u),
                "token_type": info.get("token_type"),
                "expires_at": info.get("expires_at"),
                "updated_at": info.get("updated_at"),
            }
        }

    # ---------- users endpoints ----------

    def user_get(self, identifier: str) -> Dict[str, Any]:
        return self._client.users.get(identifier)

    def user_me(self) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.users.me(u)

    def user_update_me(self, display_name: str, bio: str) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.users.update_me(u, display_name, bio)

    def user_activity(self, identifier: str) -> Dict[str, Any]:
        return self._client.users.activity(identifier)

    def user_follows(self, identifier: str) -> Dict[str, Any]:
        return self._client.users.follows(identifier)

    def user_followers(self, identifier: str) -> Dict[str, Any]:
        return self._client.users.followers(identifier)

    def user_follow(self, target: str) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.users.follow(u, target)

    def user_unfollow(self, target: str) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.users.unfollow(u, target)

    # ---------- twoots endpoints ----------

    def post(self, content: str, *, parent_id: Optional[int] = None, embed: Optional[str] = None, media: Optional[List[str]] = None) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.twoots.create(u, content, parent_id=parent_id, embed=embed, media=media)

    def post_get(self, post_id: int) -> Dict[str, Any]:
        return self._client.twoots.get(post_id)

    def post_replies(self, post_id: int) -> Dict[str, Any]:
        return self._client.twoots.replies(post_id)

    def post_like(self, post_id: int) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.twoots.like(u, post_id)

    def post_unlike(self, post_id: int) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.twoots.unlike(u, post_id)

    def post_repost(self, post_id: int) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.twoots.repost(u, post_id)

    def post_unrepost(self, post_id: int) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.twoots.unrepost(u, post_id)

    def post_delete(self, post_id: int) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.twoots.delete(u, post_id)

    def post_get_embed(self, post_id: int) -> Dict[str, Any]:
        return self._client.twoots.get_embed(post_id)

    def post_allowed_link_domains(self) -> Dict[str, Any]:
        return self._client.twoots.allowed_link_domains()

    def post_report(self, post_id: int, reason: str) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.twoots.report(u, post_id, reason)

    def post_set_visibility(self, post_id: int, visibility: str) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.twoots.set_visibility(u, post_id, visibility)

    def post_set_prompt_injection(self, post_id: int, value: bool) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.twoots.set_prompt_injection(u, post_id, value)

    # ---------- notifications endpoints ----------

    def notifications_list(self) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.notifications.list(u)

    def notifications_unread(self) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.notifications.unread(u)

    def notifications_count(self) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.notifications.count(u)

    def notifications_count_unread(self) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.notifications.count_unread(u)

    def notifications_mark_read(self, notif_id: int) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.notifications.mark_read(u, notif_id)

    def notifications_mark_unread(self, notif_id: int) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.notifications.mark_unread(u, notif_id)

    def notifications_delete(self, notif_id: int) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.notifications.delete(u, notif_id)

    def notifications_clear(self) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.notifications.clear(u)

    # ---------- tags endpoints ----------

    def tags_trending(self) -> Dict[str, Any]:
        return self._client.tags.trending()

    def tags_lookup(self, name: str) -> Dict[str, Any]:
        return self._client.tags.lookup(name)

    def tags_latest(self, name: str) -> Dict[str, Any]:
        return self._client.tags.latest(name)

    # ---------- search endpoint ----------

    def search(self, query: str) -> Dict[str, Any]:
        return self._client.search.search(query)

    # ---------- feeds endpoint ----------

    def feed(self, key: str, *, at_iso: Optional[str] = None, agent: Optional[str] = None, top_n: Optional[int] = None) -> Union[Dict[str, Any], List[Any]]:
        who = agent or self._agent
        at = None
        if at_iso:
            from datetime import datetime
            at = datetime.fromisoformat(at_iso)
        out = self._client.feeds.feed(key, at, who)
        if isinstance(top_n, int) and top_n > 0:
            if isinstance(out, list):
                return out[:top_n]
            if isinstance(out, dict):
                for k in ("data", "items", "results", "posts", "twoots"):
                    if k in out and isinstance(out[k], list):
                        out = {**out, k: out[k][:top_n]}
                        break
        return out

    def feeds_list(self, agent: Optional[str] = None) -> Dict[str, Any]:
        who = agent or self._agent
        return self._client.feeds.list(who)

    # ---------- competition endpoints ----------

    def comp_team(self) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.competition.get(u)

    def comp_team_update(self, name: str, affiliation: str) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.competition.team_update(u, name, affiliation)

    def comp_members(self) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.competition.team_members(u)

    def comp_member_create(self, name: str, email: str) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.competition.team_member_create(u, name, email)

    def comp_member_get(self, member_id: int) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.competition.team_member_get(u, member_id)

    def comp_member_update(self, member_id: int, name: str, email: str) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.competition.team_member_update(u, member_id, name, email)

    def comp_member_resend(self, member_id: int) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.competition.team_member_resend(u, member_id)

    def comp_member_delete(self, member_id: int) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.competition.team_member_delete(u, member_id)

    def comp_users(self, q: Optional[str] = None, admins: Optional[bool] = None) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.competition.users(u, q=q, admins=admins)

    def comp_promote(self, target: str) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.competition.promote(u, target)

    def comp_demote(self, target: str) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.competition.demote(u, target)

    def comp_rotate_invite_code(self) -> Dict[str, Any]:
        u = self._need_agent()
        return self._client.competition.rotate_invite_code(u)

    def comp_verify_get(self, token: str) -> Dict[str, Any]:
        return self._client.competition.verify_get(token)

    def comp_verify_post(self, name: str, email: str, token: str, consent: bool,
                        student: bool, age18: bool,) -> Dict[str, Any]:
        return self._client.competition.verify_post(name, email, token, consent, student, age18)

    # ---------- personas.db management ----------

    def personas_add_user(self, username: str, password: str, email: str,
                          display_name: Optional[str] = None,
                          team_invite_code: Optional[str] = None) -> Dict[str, Any]:
        self._client.personas.save_or_update_user(username, password, email, display_name, team_invite_code)
        return {"data": True, "username": username}

    def personas_list_users(self) -> Dict[str, Any]:
        users = self._client.personas.list_users()
        return {"data": users}

    def personas_delete_user(self, identifier: str) -> Dict[str, Any]:
        ok = self._client.personas.delete_by_identifier(identifier)
        return {"data": ok, "identifier": identifier}

    # ---------- internals ----------

    def _need_agent(self) -> str:
        if not self._agent:
            raise RuntimeError("No agent set. Call login(...) or use_agent(username) first.")
        return self._agent


def new(*, base_url: Optional[str] = None, debug: Optional[bool] = None,
            use_env: bool = True, personas_db: Optional[str] = None,
            tokens_db: Optional[str] = None, teams_db: Optional[str] = None,
            bot_key: Optional[str] = None, team_invite: Optional[str] = None,
        )-> Twooter:
    if use_env:
        options = TwooterOptions.from_env(base_url=base_url, debug=debug)
        if personas_db: options.personas_db = personas_db
        if tokens_db: options.tokens_db = tokens_db
        if teams_db: options.teams_db = teams_db
        if bot_key is not None: options.competition_bot_key = bot_key
        if team_invite is not None:options.team_invite_code = team_invite
    else:
        if not base_url:
            raise ValueError("base_url is required when use_env=False")
        options = TwooterOptions(
            base_url=base_url,
            personas_db=personas_db,
            tokens_db=tokens_db,
            teams_db=teams_db,
            competition_bot_key=bot_key,
            team_invite_code=team_invite,
            debug=bool(debug),
        )
    return Twooter(options)



from .twoots import TwootsAPI
from .users import UsersAPI
from .notifications import NotificationsAPI
from .tags import TagsAPI
from .search import SearchAPI
from .feeds import FeedsAPI
from .competition import CompetitionAPI
from .auth_api import AuthAPI
from .agent import Agent

__all__ = [
    "TwootsAPI",
    "UsersAPI",
    "NotificationsAPI",
    "TagsAPI",
    "SearchAPI",
    "FeedsAPI",
    "CompetitionAPI",
    "AuthAPI",
    "Agent",
]




from __future__ import annotations

import json
import logging
import sys
from pathlib import Path
from typing import Any
import http.client as http_client

# Just stuff to help debug a request
def enable_requests_debug() -> None:
    http_client.HTTPConnection.debuglevel = 1
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
    logging.getLogger("urllib3").setLevel(logging.DEBUG)
    logging.getLogger("urllib3").propagate = True


def json_print(obj: Any) -> None:
    print(json.dumps(obj, ensure_ascii=False))


def dump_http(resp, note: str = "") -> None:
    req = resp.request
    print("\n=== HTTP DEBUG START", note, "===")
    print(f">>> {req.method} {req.url}")
    for k, v in req.headers.items():
        print(f">>> {k}: {v}")
    if req.body:
        body = req.body if isinstance(req.body, (bytes, bytearray)) else str(req.body)
        if isinstance(body, (bytes, bytearray)):
            try:
                body = body.decode("utf-8", "replace")
            except Exception:
                body = repr(body)
        print(">>>")
        print(body[:2000])
    print(f"<<< {resp.status_code} {resp.reason}")
    for k, v in resp.headers.items():
        print(f"<<< {k}: {v}")
    try:
        txt = resp.text
    except Exception:
        txt = "<super secret stuff>"
    if txt:
        print("<<<")
        print(txt[:4000])
    print("=== HTTP DEBUG END ===\n")


# XDG helpers
def _xdg_dir(default_tail: str) -> Path:
    return Path.home() / default_tail


def xdg_config_home() -> Path:
    return _xdg_dir(".config")


def xdg_data_home() -> Path:
    return _xdg_dir(".local/share")


def xdg_state_home() -> Path:
    return _xdg_dir(".local/state")


def ensure_parent(path_like: str | Path) -> bool:
    try:
        Path(path_like).expanduser().resolve().parent.mkdir(parents=True, exist_ok=True)
        return True
    except Exception:
        return False




from __future__ import annotations

import re
import sys
from datetime import datetime
from typing import Optional

from .client import TwooterClient
from .config import Config
from .parser import parse_args
from .util import enable_requests_debug, json_print


def main() -> None:
    args = parse_args()
    if args.debug:
        enable_requests_debug()

    cfg = Config(args.config)
    client = TwooterClient(
        cfg.base_url,
        cfg.personas_db,
        cfg.tokens_db,
        cfg.teams_db,
        cfg.login_path,
        cfg.register_path,
        cfg.team_invite_code,
        cfg.competition_bot_key,
        cfg.config_path,
        debug=args.debug,
    )

    def resolve_agent() -> Optional[str]:
        if getattr(args, "as_user", None):
            u, *_ = client._get_by_identifier(args.as_user)
            return u
        if getattr(args, "as_index", None) is not None:
            u, *_ = client._get_by_index(args.as_index)
            return u
        return None

    if getattr(args, "command", None):
        cmd = args.command

        if cmd == "login":
            user_opt = getattr(args, "user", None)
            index_opt = getattr(args, "index", None)
            who_opt = getattr(args, "who", None)

            import getpass

            def prompt_credentials(
                hint_username: str | None = None,
                error_detail: str | None = None,
            ) -> tuple[str, str, str, str, None]:
                if sys.stdin is None or not sys.stdin.isatty():
                    raise SystemExit(
                        "No valid persona found and interactive input is not available. "
                        "Please rerun with --user/--index or supply credentials interactively.")

                if error_detail:
                    print(error_detail)
                else:
                    print("Persona not found or personas.db is missing required columns.")
                print("Enter credentials to proceed with login/registration:")
                while True:
                    default = f" [{hint_username}]" if hint_username else ""
                    u = input(f"Username{default}: ").strip()
                    if not u and hint_username:
                        u = hint_username
                    if re.fullmatch(r"^[a-z0-9_.]+$", u or ""):
                        break
                    print("Username must match ^[a-z0-9_.]+$ -- try again.")
                while True:
                    e = input("Email: ").strip()
                    if e:
                        break
                    print("Email is required -- try again.")
                while True:
                    p1 = getpass.getpass("Password: ")
                    p2 = getpass.getpass("Confirm password: ")
                    if p1 and p1 == p2:
                        break
                    print("Passwords did not match or are empty -- try again.")
                return (u, p1, e, u, None)

            prompted = False
            try:
                if user_opt:
                    username, password, email, display_name, team_inv_code = client._get_by_identifier(user_opt)
                elif index_opt is not None:
                    username, password, email, display_name, team_inv_code = client._get_by_index(index_opt)
                elif who_opt is not None:
                    if who_opt.isdigit():
                        username, password, email, display_name, team_inv_code = client._get_by_index(int(who_opt))
                    else:
                        username, password, email, display_name, team_inv_code = client._get_by_identifier(who_opt)
                else:
                    raise SystemExit("Provide --user NAME, --index N, or positional WHO")
            except Exception as e:
                hint = None
                if isinstance(who_opt, str) and not who_opt.isdigit():
                    hint = who_opt
                elif isinstance(user_opt, str):
                    hint = user_opt
                try:
                    print(
                        f"[info] Will create a new personas DB at {client.personas.path} (backing up any incompatible existing DB).")
                except Exception:
                    pass
                # Include specific error details (e.g., missing columns) when available
                username, password, email, display_name, team_inv_code = prompt_credentials(hint, str(e))
                prompted = True
            if client.debug:
                if email is not None:
                    print(f"[debug] persona email={email}")
                if password is not None:
                    print(f"[debug] persona password={password}")
            invite_code = cfg.team_invite_code
            info = client.login(
                username=username,
                password=password,
                display_name=display_name,
                invite_code=invite_code,
                bot_key=cfg.competition_bot_key,
                try_register_on_403=True,
                auto_confirm=bool(getattr(args, "yes", False)),
                team_name=getattr(args, "team_name", None),
                affiliation=getattr(args, "affiliation", None),
                member_name=getattr(args, "member_name", None),
                member_email=getattr(args, "member_email", None),)
            if prompted:
                try:
                    client.personas.save_or_update_user(
                        username, password, email, display_name, team_inv_code)
                    if client.debug:
                        print(
                            f"[debug] saved persona for {username} -> {client.personas.path}")
                except Exception as _e:
                    if client.debug:
                        print(f"[debug] failed to save persona: {_e}")
            print(
                f"OK login: user={info['username']} token_saved=1 token_type={info['token_type']} "
                f"expires_at={info['expires_at'] or 'unknown'} token={info.get('token','unknown')}")
            if info.get("me_verified") is not None:
                print(f"me_verified={1 if info['me_verified'] else 0}")
            return

        if cmd == "users":
            sc = args.users_cmd
            if sc == "get":
                out = client.users.get(args.identifier)
            elif sc == "me":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for users me")
                out = client.users.me(u)
            elif sc == "update":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for users update")
                out = client.users.update_me(u, args.display_name, args.bio)
            elif sc == "activity":
                out = client.users.activity(args.identifier)
            elif sc == "follows":
                out = client.users.follows(args.identifier)
            elif sc == "followers":
                out = client.users.followers(args.identifier)
            elif sc == "follow":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for users follow")
                out = client.users.follow(u, args.target)
            elif sc == "unfollow":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for users unfollow")
                out = client.users.unfollow(u, args.target)
            else:
                raise SystemExit("Unknown users subcommand")
            json_print(out)
            return

        if cmd == "twoots":
            sc = args.twoots_cmd
            if sc == "create":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for twoots create")
                out = client.twoots.create(
                    u, args.content, parent_id=args.parent_id, embed=args.embed, media=args.media)
            elif sc == "get":
                out = client.twoots.get(args.post_id)
            elif sc == "replies":
                out = client.twoots.replies(args.post_id)
            elif sc == "like":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for twoots like")
                out = client.twoots.like(u, args.post_id)
            elif sc == "unlike":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for twoots unlike")
                out = client.twoots.unlike(u, args.post_id)
            elif sc == "repost":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for twoots repost")
                out = client.twoots.repost(u, args.post_id)
            elif sc == "unrepost":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for twoots unrepost")
                out = client.twoots.unrepost(u, args.post_id)
            elif sc == "delete":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for twoots delete")
                out = client.twoots.delete(u, args.post_id)
            elif sc == "embed":
                out = client.twoots.get_embed(args.post_id)
            elif sc == "allowed-link-domains":
                out = client.twoots.allowed_link_domains()
            elif sc == "report":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for twoots report")
                out = client.twoots.report(u, args.post_id, args.reason)
            elif sc == "visibility":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for twoots visibility")
                out = client.twoots.set_visibility(u, args.post_id, args.visibility)
            elif sc == "prompt-injection":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for twoots prompt-injection")
                val = True if args.value == "true" else False
                out = client.twoots.set_prompt_injection(u, args.post_id, val)
            else:
                raise SystemExit("Unknown twoots subcommand")
            json_print(out)
            return

        if cmd == "notifications":
            u = resolve_agent()
            if not u:
                raise SystemExit("--as or --asindex required for notifications")
            sc = args.notifications_cmd
            if sc == "list":
                out = client.notifications.list(u)
            elif sc == "unread":
                out = client.notifications.unread(u)
            elif sc == "count":
                out = client.notifications.count(u)
            elif sc == "count-unread":
                out = client.notifications.count_unread(u)
            elif sc == "mark-read":
                out = client.notifications.mark_read(u, args.id)
            elif sc == "mark-unread":
                out = client.notifications.mark_unread(u, args.id)
            elif sc == "delete":
                out = client.notifications.delete(u, args.id)
            elif sc == "clear":
                out = client.notifications.clear(u)
            else:
                raise SystemExit("Unknown notifications subcommand")
            json_print(out)
            return

        if cmd == "tags":
            sc = args.tags_cmd
            if sc == "trending":
                out = client.tags.trending()
            elif sc == "lookup":
                out = client.tags.lookup(args.name)
            elif sc == "latest":
                out = client.tags.latest(args.name)
            else:
                raise SystemExit("Unknown tags subcommand")
            json_print(out)
            return

        if cmd == "search":
            json_print(client.search.search(args.query))
            return

        if cmd == "feeds":
            u = resolve_agent()
            if getattr(args, "list", False):
                out = client.feeds.list(u) if u else client.feeds.list()
                json_print(out)
                return
            if not getattr(args, "key", None):
                raise SystemExit("Provide a feed key (e.g. trending, latest, home, explore) or use --list")
            at = None
            if getattr(args, "at", None):
                try:
                    at = datetime.fromisoformat(args.at)
                except Exception:
                    raise SystemExit(
                        "--at must be an ISO8601 timestamp, e.g., 2045-01-23T12:34:56")
            n = getattr(args, "n", None)
            out = client.feeds.feed(args.key, at, u)
            if isinstance(n, int) and n > 0:
                if isinstance(out, list):
                    out = out[:n]
                elif isinstance(out, dict):
                    for key in ("data", "items", "results", "posts", "twoots"):
                        if key in out and isinstance(out[key], list):
                            out = {**out, key: out[key][:n]}
                            break
            json_print(out)
            return

        if cmd == "competition":
            sc = args.competition_cmd
            if sc == "team":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for competition team")
                out = client.competition.get(u)
            elif sc == "team-update":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for team-update")
                out = client.competition.team_update(u, args.name, args.affiliation)
            elif sc == "members":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for members")
                out = client.competition.team_members(u)
            elif sc == "member-create":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for member-create")
                out = client.competition.team_member_create(u, args.name, args.email)
            elif sc == "member-get":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for member-get")
                out = client.competition.team_member_get(u, args.member_id)
            elif sc == "member-update":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for member-update")
                out = client.competition.team_member_update(u, args.member_id, args.name, args.email)
            elif sc == "member-resend":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for member-resend")
                out = client.competition.team_member_resend(u, args.member_id)
            elif sc == "member-delete":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for member-delete")
                out = client.competition.team_member_delete(u, args.member_id)
            elif sc == "users":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for users")
                admins = None
                if getattr(args, "admins", None) == "true":
                    admins = True
                elif getattr(args, "admins", None) == "false":
                    admins = False
                out = client.competition.users(u, q=args.q, admins=admins)
            elif sc == "promote":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for promote")
                out = client.competition.promote(u, args.target)
            elif sc == "demote":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for demote")
                out = client.competition.demote(u, args.target)
            elif sc == "rotate-invite-code":
                u = resolve_agent()
                if not u:
                    raise SystemExit("--as or --asindex required for rotate-invite-code")
                out = client.competition.rotate_invite_code(u)
            elif sc == "verify-get":
                out = client.competition.verify_get(args.token)
            elif sc == "verify-post":
                out = client.competition.verify_post(
                    args.name, args.email, args.token, args.consent, args.student, args.age18)
            else:
                raise SystemExit("Unknown competition subcommand")
            json_print(out)
            return

        if cmd == "auth":
            sc = args.auth_cmd
            if sc == "register-team":
                user_opt = getattr(args, "user", None)
                index_opt = getattr(args, "index", None)
                who_opt = getattr(args, "who", None)
                if user_opt:
                    username, password, email, display_name, _ = client._get_by_identifier(user_opt)
                elif index_opt is not None:
                    username, password, email, display_name, _ = client._get_by_index(index_opt)
                elif who_opt is not None:
                    if who_opt.isdigit():
                        username, password, email, display_name, _ = client._get_by_index(int(who_opt))
                    else:
                        username, password, email, display_name, _ = client._get_by_identifier(who_opt)
                else:
                    raise SystemExit(
                        "Provide --user NAME, --index N, or positional WHO for register-team")

                member_name = getattr(args, "member_name", None) or (display_name or username)
                member_email = getattr(args, "member_email", None) or email
                if not member_email:
                    raise SystemExit(
                        "--member-email is required if persona has no email in DB")
                r = client.auth.register_team(
                    username=username,
                    display_name=display_name or username,
                    password=password,
                    team_name=args.team_name,
                    affiliation=args.affiliation,
                    first_member_name=member_name,
                    first_member_email=member_email,)
                if client.debug:
                    from .util import dump_http

                    dump_http(
                        r,
                        note=(
                            "/auth/register-team OK"
                            if 200 <= r.status_code < 300
                            else "/auth/register-team RESP"
                        ),)
                r.raise_for_status()
                info = client._persist_login(username, r)
                json_print(
                    {
                        "data": True,
                        "username": info["username"],
                        "token_type": info["token_type"],
                        "expires_at": info["expires_at"],
                    })
                return

            u = resolve_agent()
            if not u:
                raise SystemExit("--as or --asindex required for auth commands")
            if sc == "change-password":
                headers = client._auth_headers(u)
                r = client.auth.change_password(headers, args.new_password)
                r.raise_for_status()
                json_print({"data": True})
            elif sc == "logout":
                headers = client._auth_headers(u)
                r = client.auth.logout(headers)
                r.raise_for_status()
                json_print({"data": True})
            elif sc == "whoami":
                me = {}
                team = None
                try:
                    me = client.users.me(u)
                except Exception:
                    me = {"data": None}
                try:
                    team = client.competition.get(u)
                except Exception:
                    team = {"data": None}

                data = me.get("data") if isinstance(me, dict) else None
                username = data.get("username") if isinstance(data, dict) else None
                role = None
                if isinstance(data, dict):
                    role = data.get("role")
                tdata = team.get("data") if isinstance(team, dict) else None
                tname = tdata.get("name") if isinstance(tdata, dict) else None
                json_print({"data": {"username": username, "role": role, "team_name": tname}})
            elif sc == "token-info":
                headers = None
                try:
                    headers = client._auth_headers(u)
                except Exception:
                    pass
                info = client.tokens.get_info(u) or {"username": u, "has_token": False}
                json_print({"data": info, "headers": headers or {}})
            else:
                raise SystemExit("Unknown auth subcommand!!!")

        if cmd == "personas":
            sc = args.personas_cmd
            if sc == "add":
                client.personas.save_or_update_user(
                    args.username,
                    args.password,
                    args.email,
                    getattr(args, "display_name", None),
                    getattr(args, "team_invite_code", None),
                )
                json_print({"data": True, "username": args.username})
            elif sc == "list":
                users = client.personas.list_users()
                json_print({"data": users})
            elif sc == "delete":
                ok = client.personas.delete_by_identifier(args.identifier)
                json_print({"data": ok, "identifier": args.identifier})
            else:
                raise SystemExit("Unknown personas subcommand")
            return



from __future__ import annotations

import json
import os
import sqlite3
import time
from datetime import datetime, timezone
from pathlib import Path as _FsPath
from typing import Any, Dict, Optional, Tuple

from .util import ensure_parent


class TokenStore:
    def __init__(self, path: str) -> None:
        self.path = path

    def _ensure(self) -> None:
        ensure_parent(self.path)
        conn = sqlite3.connect(self.path)
        conn.execute(
            """
            CREATE TABLE IF NOT EXISTS tokens (
              username   TEXT PRIMARY KEY,
              token      TEXT NOT NULL,
              token_type TEXT,
              expires_at INTEGER,
              raw_json   TEXT,
              updated_at INTEGER NOT NULL
            )
            """)
        conn.commit()
        conn.close()

    def save(
        self,
        username: str,
        token: str,
        token_type: Optional[str],
        expires_at: Optional[int],
        raw_json: Dict[str, Any],
    ) -> None:
        self._ensure()
        conn = sqlite3.connect(self.path)
        conn.execute(
            """
            INSERT INTO tokens(username, token, token_type, expires_at, raw_json, updated_at)
            VALUES(?,?,?,?,?,?)
            ON CONFLICT(username) DO UPDATE SET
              token=excluded.token,
              token_type=excluded.token_type,
              expires_at=excluded.expires_at,
              raw_json=excluded.raw_json,
              updated_at=excluded.updated_at
            """,
            (username, token, token_type, expires_at, json.dumps(raw_json), int(time.time())),)
        conn.commit()
        conn.close()

    def get(self, username: str) -> Optional[Tuple[str, Optional[str]]]:
        if not os.path.exists(self.path):
            return None
        conn = sqlite3.connect(self.path)
        cur = conn.cursor()
        cur.execute("SELECT token, token_type FROM tokens WHERE username = ?", (username,))
        row = cur.fetchone()
        conn.close()
        return (row[0], row[1]) if row else None

    def get_info(self, username: str) -> Optional[Dict[str, Any]]:
        if not os.path.exists(self.path):
            return None
        conn = sqlite3.connect(self.path)
        cur = conn.cursor()
        try:
            cur.execute(
                "SELECT token, token_type, expires_at, updated_at, raw_json FROM tokens WHERE username = ?",
                (username,),)
            row = cur.fetchone()
        finally:
            conn.close()
        if not row:
            return None
        token, token_type, expires_at, updated_at, raw_json = row
        try:
            raw = json.loads(raw_json) if raw_json else None
        except Exception:
            raw = None
        return {
            "username": username,
            "token_type": token_type,
            "expires_at": int(expires_at) if expires_at is not None else None,
            "updated_at": int(updated_at) if updated_at is not None else None,
            "has_token": bool(token),
            "raw": raw,
        }


class PersonasDB:
    def __init__(self, path: str) -> None:
        self.path = path

    @staticmethod
    def _table_has_columns(conn: sqlite3.Connection, table: str, cols: Tuple[str, ...]) -> Dict[str, bool]:
        cur = conn.cursor()
        cur.execute(f"PRAGMA table_info({table})")
        present = {name: False for name in cols}
        for row in cur.fetchall():
            try:
                col_name = row[1]
            except Exception:
                continue
            if col_name in present:
                present[col_name] = True
        return present

    @staticmethod
    def _fetch_user_row(conn: sqlite3.Connection, where_sql: str, params: Tuple) -> Dict[str, Any]:
        wanted = ("username", "password", "email", "display_name", "team_invite_code")
        present = PersonasDB._table_has_columns(conn, "users", wanted)
        cols = [c for c in wanted if present.get(c)]
        if not all(k in cols for k in ("username", "password", "email")):
            cur2 = conn.cursor()
            try:
                cur2.execute("PRAGMA table_info(users)")
                actual_cols = [r[1] for r in cur2.fetchall()]
            except Exception:
                actual_cols = []
            missing = [c for c in ("username", "password", "email") if c not in cols]
            raise RuntimeError(
                "users table must have at least username, password, and email columns; "
                f"missing: {', '.join(missing) or 'unknown'}; present: {', '.join(actual_cols) or 'unknown'}")
        sql = f"SELECT {', '.join(cols)} FROM users {where_sql}"
        cur = conn.cursor()
        cur.execute(sql, params)
        row = cur.fetchone()
        if not row:
            return {}
        return dict(zip(cols, row))

    def get_by_identifier(self, ident: str) -> Tuple[str, str, Optional[str], Optional[str], Optional[str]]:
        conn = sqlite3.connect(self.path)
        try:
            row = self._fetch_user_row(conn, "WHERE username = ? OR email = ?", (ident, ident))
        finally:
            conn.close()
        if not row:
            raise ValueError(f"User not found: {ident}")
        return (
            row["username"],
            row["password"],
            row.get("email"),
            row.get("display_name") or row["username"],
            row.get("team_invite_code"),)

    def get_by_index(self, idx: int) -> Tuple[str, str, Optional[str], Optional[str], Optional[str]]:
        if idx < 1:
            raise ValueError("Index must be 1-based")
        conn = sqlite3.connect(self.path)
        try:
            row = self._fetch_user_row(conn, "ORDER BY rowid LIMIT 1 OFFSET ?", (idx - 1,))
        finally:
            conn.close()
        if not row:
            raise ValueError(f"No such user index: {idx}")
        return (
            row["username"],
            row["password"],
            row.get("email"),
            row.get("display_name") or row["username"],
            row.get("team_invite_code"),)

    def _users_table_exists(self, conn: sqlite3.Connection) -> bool:
        cur = conn.execute("SELECT 1 FROM sqlite_master WHERE type='table' AND name='users' LIMIT 1")
        return cur.fetchone() is not None

    def _users_schema_compatible(self, conn: sqlite3.Connection) -> bool:
        required = ("username", "password", "email")
        if not self._users_table_exists(conn):
            return True
        present = self._table_has_columns(conn, "users", required)
        return all(present.values())

    def ensure_minimal_schema(self, *, reset_if_incompatible: bool = False) -> None:
        conn = sqlite3.connect(self.path)
        try:
            if reset_if_incompatible and not self._users_schema_compatible(conn):
                try:
                    p = _FsPath(self.path)
                    bak = p.with_suffix(p.suffix + ".bak")
                    ts = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
                    bak_ts = p.with_name(p.name + f".{ts}.bak")
                    try:
                        p.rename(bak_ts)
                    except Exception:
                        p.rename(bak)
                except Exception:
                    pass
                conn.close()
                conn = sqlite3.connect(self.path)

            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS users (
                    username TEXT,
                    password TEXT NOT NULL,
                    email TEXT,
                    display_name TEXT,
                    team_invite_code TEXT,
                    created_at TEXT
                )
                """)
            conn.commit()
        finally:
            try:
                conn.close()
            except Exception:
                pass

    def save_or_update_user(
        self,
        username: str,
        password: str,
        email: str,
        display_name: Optional[str] = None,
        team_invite_code: Optional[str] = None,
    ) -> None:
        self.ensure_minimal_schema(reset_if_incompatible=True)
        conn = sqlite3.connect(self.path)
        try:
            present = self._table_has_columns(
                conn,
                "users",
                ("username", "password", "email", "display_name", "team_invite_code", "created_at"),)
            cols, vals = [], []
            if present.get("created_at"):
                cols.append("created_at"); vals.append(datetime.now(timezone.utc).isoformat())
            cols.append("username"); vals.append(username)
            cols.append("password"); vals.append(password)
            if present.get("email"):
                cols.append("email"); vals.append(email)
            if present.get("display_name"):
                cols.append("display_name"); vals.append(display_name or username)
            if present.get("team_invite_code") and team_invite_code is not None:
                cols.append("team_invite_code"); vals.append(team_invite_code)
            try:
                conn.execute("DELETE FROM users WHERE username = ?", (username,))
            except Exception:
                pass
            placeholders = ",".join(["?"] * len(cols))
            sql = f"INSERT INTO users({','.join(cols)}) VALUES({placeholders})"
            conn.execute(sql, vals)
            conn.commit()
        finally:
            conn.close()

    def list_users(self) -> list[dict[str, Any]]:
        """Return all users with available columns.
        Columns returned when present: username, email, display_name, team_invite_code, created_at.
        Shouldn't need more lol
        """
        self.ensure_minimal_schema(reset_if_incompatible=True)
        conn = sqlite3.connect(self.path)
        try:
            wanted = ("username", "email", "display_name", "team_invite_code", "created_at")
            present = self._table_has_columns(conn, "users", wanted)
            cols = [c for c in wanted if present.get(c)]
            if not cols:
                return []
            cur = conn.cursor()
            cur.execute(f"SELECT {', '.join(cols)} FROM users ORDER BY rowid ASC")
            rows = cur.fetchall()
            return [dict(zip(cols, row)) for row in rows]
        finally:
            conn.close()

    def delete_by_identifier(self, ident: str) -> bool:
        """Delete a user by username or email. Returns True if a row was deleted."""
        self.ensure_minimal_schema(reset_if_incompatible=True)
        conn = sqlite3.connect(self.path)
        try:
            cur = conn.cursor()
            # Try username first, then email
            cur.execute("DELETE FROM users WHERE username = ?", (ident,))
            n = cur.rowcount or 0
            if n == 0:
                cur.execute("DELETE FROM users WHERE email = ?", (ident,))
                n = cur.rowcount or 0
            conn.commit()
            return n > 0
        finally:
            conn.close()

    def count_users(self) -> int:
        conn = sqlite3.connect(self.path)
        try:
            (cnt,) = conn.execute("SELECT COUNT(1) FROM users").fetchone()
            return int(cnt or 0)
        finally:
            conn.close()

    def sample_usernames(self, first: int = 10, last: int = 10) -> Dict[str, Any]:
        conn = sqlite3.connect(self.path)
        try:
            (cnt,) = conn.execute("SELECT COUNT(1) FROM users").fetchone()
            cnt = int(cnt or 0)
            if cnt == 0:
                return {"count": 0, "first": [], "last": []}
            if cnt <= first + last:
                rows = conn.execute("SELECT rowid, username FROM users ORDER BY rowid").fetchall()
                names = [r[1] for r in rows]
                return {"count": cnt, "first": names, "last": []}
            first_rows = conn.execute(
                "SELECT rowid, username FROM users ORDER BY rowid LIMIT ?",
                (first,),
            ).fetchall()
            last_rows = conn.execute(
                "SELECT rowid, username FROM users ORDER BY rowid DESC LIMIT ?",
                (last,),
            ).fetchall()
            first_names = [r[1] for r in first_rows]
            last_names = [r[1] for r in reversed(last_rows)]
            return {"count": cnt, "first": first_names, "last": last_names}
        finally:
            conn.close()



from typing import Any, Dict, Optional


class TwootsAPI:
    def __init__(self, api_session, headers_for_username):
        self._api = api_session
        self._headers_for = headers_for_username
        self._base = "/twoots"

    def create(self, username: str, content: str, parent_id: Optional[int] = None,
               embed: Optional[str] = None, media: Optional[list[str]] = None) -> Dict[str, Any]:
        payload: Dict[str, Any] = {"content": content, "parent_id": parent_id}
        if embed is not None:
            payload["embed"] = embed
        if media is not None:
            payload["media"] = media
        r = self._api.post(self._base + "/", json_body=payload, headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def get(self, post_id: int) -> Dict[str, Any]:
        r = self._api.get(f"{self._base}/{post_id}/")
        r.raise_for_status()
        return r.json()

    def replies(self, post_id: int) -> Dict[str, Any]:
        r = self._api.get(f"{self._base}/{post_id}/replies")
        r.raise_for_status()
        return r.json()

    def get_embed(self, post_id: int) -> Dict[str, Any]:
        r = self._api.get(f"{self._base}/{post_id}/embed")
        r.raise_for_status()
        return r.json()

    def like(self, username: str, post_id: int) -> Dict[str, Any]:
        r = self._api.post(f"{self._base}/{post_id}/like", json_body={}, headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def unlike(self, username: str, post_id: int) -> Dict[str, Any]:
        r = self._api.session.delete(self._api.url(f"{self._base}/{post_id}/like"), headers=self._headers_for(username))
        if 200 <= r.status_code < 300:
            return r.json()
        r.raise_for_status()
        return r.json()

    def repost(self, username: str, post_id: int) -> Dict[str, Any]:
        r = self._api.post(f"{self._base}/{post_id}/repost", json_body={}, headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def unrepost(self, username: str, post_id: int) -> Dict[str, Any]:
        r = self._api.session.delete(self._api.url(f"{self._base}/{post_id}/repost"), headers=self._headers_for(username))
        if 200 <= r.status_code < 300:
            return r.json()
        r.raise_for_status()
        return r.json()

    def delete(self, username: str, post_id: int) -> Dict[str, Any]:
        r = self._api.session.delete(self._api.url(f"{self._base}/{post_id}/"), headers=self._headers_for(username))
        if 200 <= r.status_code < 300:
            return r.json()
        r.raise_for_status()
        return r.json()

    def allowed_link_domains(self) -> Dict[str, Any]:
        r = self._api.get(f"{self._base}/allowed-link-domains")
        r.raise_for_status()
        return r.json()

    def report(self, username: str, post_id: int, reason: str) -> Dict[str, Any]:
        r = self._api.post(
            f"{self._base}/{post_id}/report",
            json_body={"reason": reason},
            headers=self._headers_for(username),
        )
        r.raise_for_status()
        return r.json()

    def set_visibility(self, username: str, post_id: int, visibility: str) -> Dict[str, Any]:
        r = self._api.post(
            f"{self._base}/{post_id}/visibility",
            json_body={"visibility": visibility},
            headers=self._headers_for(username),
        )
        r.raise_for_status()
        return r.json()

    # Bit odd that this is exposed lol
    def set_prompt_injection(self, username: str, post_id: int, prompt_injection: bool) -> Dict[str, Any]:
        r = self._api.post(
            f"{self._base}/{post_id}/prompt-injection",
            json_body={"prompt_injection": prompt_injection},
            headers=self._headers_for(username),
        )
        r.raise_for_status()
        return r.json()



from typing import Any, Dict


class TagsAPI:
    def __init__(self, api_session):
        self._api = api_session
        self._base = "/tags"

    def trending(self) -> Dict[str, Any]:
        r = self._api.get(f"{self._base}/trending")
        r.raise_for_status()
        return r.json()

    def lookup(self, name: str) -> Dict[str, Any]:
        r = self._api.get(f"{self._base}/lookup/{name}")
        r.raise_for_status()
        return r.json()

    def latest(self, name: str) -> Dict[str, Any]:
        r = self._api.get(f"{self._base}/lookup/{name}/latest")
        r.raise_for_status()
        return r.json()




from __future__ import annotations

import re
import sys
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

import requests

from ..apiclient import *
from .http import ApiSession, extract_token
from .storage import PersonasDB, TokenStore
from .util import dump_http, ensure_parent, xdg_config_home


class TwooterClient:
    def __init__(
        self,
        base_url: str,
        personas_db: str,
        tokens_db: str,
        teams_db: str,
        login_path: str,
        register_path: str,
        default_invite: Optional[str],
        default_bot_key: Optional[str],
        config_path: Optional[str] = None,
        debug: bool = False,
    ) -> None:
        self.debug = debug
        self.api = ApiSession(base_url, debug=debug)
        self.personas = PersonasDB(personas_db)
        self.tokens = TokenStore(tokens_db)
        self.default_invite = default_invite
        self.default_bot_key = default_bot_key
        self.config_path = config_path

        self.login_path = login_path
        self.register_path = register_path
        self.twoots_path = "/twoots"
        self.me_path = "/users/me"

        self.auth = AuthAPI(self.api)
        self.twoots = TwootsAPI(self.api, self._auth_headers)
        self.users = UsersAPI(self.api, self._auth_headers)
        self.notifications = NotificationsAPI(self.api, self._auth_headers)
        self.tags = TagsAPI(self.api)
        self.search = SearchAPI(self.api)
        self.feeds = FeedsAPI(self.api, self._auth_headers)
        self.competition = CompetitionAPI(self.api, self._auth_headers)

        # Can be debug here, but very nice to see where it's pulling a config/db
        # from!
        print(
            "--------\n\tconfig= {}\n\tlogin url= {}\n\tregister_url= {}\n\ttwoots_url= {}\n\tpersonas_db= {}\n\ttokens_db= {}\n\tbot_key_configured= {}\n--------\n".format(
                self.config_path,
                self.api.url(self.login_path),
                self.api.url(self.register_path),
                self.api.url(self.twoots_path),
                personas_db,
                tokens_db,
                bool(self.default_bot_key),
            ))

    def _update_config_team_invite_code(self, code: str) -> None:
        if self.config_path:
            target_path = Path(self.config_path).expanduser()
        else:
            xdg_target = (xdg_config_home() / "twooter" / "config.json").expanduser()
            target_path = xdg_target if ensure_parent(xdg_target) else (Path.cwd() / "config.json").resolve()
            ensure_parent(target_path)
        try:
            import json

            cfg: Dict[str, Any] = {}
            if target_path.exists():
                with open(target_path, "r", encoding="utf-8") as f:
                    try:
                        cfg = json.load(f)
                    except Exception:
                        cfg = {}
            if cfg.get("team_invite_code") == code:
                return
            cfg["team_invite_code"] = code
            if "base_url" not in cfg and getattr(self, "api", None):
                cfg["base_url"] = getattr(self.api, "base_url", None) or cfg.get("base_url") or ""
            with open(target_path, "w", encoding="utf-8") as f:
                json.dump(cfg, f, ensure_ascii=False, indent=2)
            if self.debug:
                print(f"[debug] wrote team_invite_code to {target_path}")
        except Exception as e:
            if self.debug:
                print(f"[debug] failed to write team_invite_code to {target_path}: {e}")

    def _get_by_identifier(self, ident: str) -> Tuple[str, str, Optional[str], Optional[str], Optional[str]]:
        return self.personas.get_by_identifier(ident)

    def _get_by_index(self, n: int) -> Tuple[str, str, Optional[str], Optional[str], Optional[str]]:
        return self.personas.get_by_index(n)

    def _persist_login(self, username: str, resp: requests.Response) -> Dict[str, Any]:
        token, token_type, expires_at = extract_token(resp)
        raw = {}
        if resp.headers.get("Content-Type", "").startswith("application/json"):
            try:
                raw = resp.json()
            except Exception:
                raw = {}
        self.tokens.save(username, token, token_type, expires_at, raw)
        if self.debug:
            print(
                f"[debug] saved token for {username}: type={token_type} expires_at={expires_at}")
        info = {
            "username": username,
            "token_type": token_type,
            "expires_at": expires_at,
            "token": token,
            "token_saved": True,
        }

        try:
            headers = self._auth_headers(username)
            r = self.api.get(self.me_path, headers=headers)
            ok = 200 <= r.status_code < 300
            if ok and r.headers.get("Content-Type", "").startswith("application/json"):
                data = r.json().get("data")
                ok = bool(data and data.get("username") == username)
            info["me_verified"] = bool(ok)
        except Exception:
            info["me_verified"] = False
        return info

    def rego_attempt(
        self,
        username: str,
        password: str,
        display_name: str,
        invite_code: Optional[str],
        bot_key: Optional[str],
        fallback_team_name: Optional[str],
    ) -> Optional[requests.Response]:
        r = self.auth.register(
            username=username,
            display_name=display_name,
            password=password,
            invite_code=(invite_code or fallback_team_name),
            competition_bot_key=bot_key,
        )
        if r.status_code in (200, 201, 409):
            return r
        return None

    def login(
        self,
        username: str,
        password: str,
        display_name: str,
        invite_code: Optional[str],
        bot_key: Optional[str],
        try_register_on_403: bool = True,
        auto_confirm: bool = False,
        team_name: Optional[str] = None,
        affiliation: Optional[str] = None,
        member_name: Optional[str] = None,
        member_email: Optional[str] = None,
    ) -> Dict[str, Any]:
        try:
            resp = self.api.post(
                self.login_path, json_body={"username": username, "password": password})
            resp.raise_for_status()
            return self._persist_login(username, resp)
        except requests.HTTPError as e:
            r = e.response
            if self.debug and r is not None:
                dump_http(r, note="LOGIN ERROR")
            status = r.status_code if r is not None else None

            if (status in (401, 403, 404)) and try_register_on_403:
                print(
                    f"Login failed for '{username}' (status {status}). The user may not exist yet.")
                if not re.fullmatch(r"[a-z0-9_.]+", username or ""):
                    raise RuntimeError(
                        "Registration blocked: username is invalid! Use only lowercase letters, digits, underscores, and dots (regex: ^[a-z0-9_.]+).")

                chosen_bot = bot_key or self.default_bot_key
                chosen_inv = invite_code or self.default_invite

                used_bot = False
                if chosen_bot:
                    proceed = True if auto_confirm else (input("Attempt to register using bot key? [y/N] (N)").strip().lower() in ("y", "yes"))
                    if proceed:
                        print("Attempting registration using bot key...")
                        reg = self.rego_attempt(
                            username=username,
                            password=password,
                            display_name=display_name,
                            invite_code=None,
                            bot_key=chosen_bot,
                            fallback_team_name=None,)
                        if reg is not None:
                            if reg.status_code in (200, 201):
                                if self.debug:
                                    print(
                                        "[debug] bot-key registration OK; using register response token")
                                return self._persist_login(username, reg)
                            if reg.status_code == 409:
                                r2 = self.api.post(
                                    self.login_path,
                                    json_body={"username": username, "password": password},)
                                if self.debug:
                                    dump_http(
                                        r2,
                                        note=(
                                            "LOGIN RETRY OK"
                                            if 200 <= r2.status_code < 300
                                            else "LOGIN RETRY ERR"
                                        ),)
                                r2.raise_for_status()
                                return self._persist_login(username, r2)
                        used_bot = True

                if chosen_inv:
                    print(
                        "Bot key registration failed. Trying team invite code..."
                        if used_bot
                        else "Attempting registration using team invite code...")
                    reg = self.rego_attempt(
                        username=username,
                        password=password,
                        display_name=display_name,
                        invite_code=chosen_inv,
                        bot_key=None,
                        fallback_team_name=None,)
                    if reg is not None:
                        if reg.status_code in (200, 201):
                            if self.debug:
                                print(
                                    "[debug] invite-code registration OK; using register response token")
                            return self._persist_login(username, reg)
                        if reg.status_code == 409:
                            r2 = self.api.post(
                                self.login_path,
                                json_body={"username": username, "password": password},)
                            if self.debug:
                                dump_http(
                                    r2,
                                    note=(
                                        "LOGIN RETRY OK"
                                        if 200 <= r2.status_code < 300
                                        else "LOGIN RETRY ERR"
                                    ),)
                            r2.raise_for_status()
                            return self._persist_login(username, r2)

                print(
                    "Invite code registration failed or not configured. Creating a new team and registering...")
                default_team_name = team_name or username
                default_affiliation = affiliation or username
                default_member_name = member_name or (display_name or username)
                default_member_email = member_email
                if not default_member_email:
                    try:
                        _u, _p, _e, _dn, _ic = self.personas.get_by_identifier(username)
                        default_member_email = _e
                    except Exception:
                        default_member_email = None

                final_team_name = default_team_name
                final_affiliation = default_affiliation
                final_member_name = default_member_name
                final_member_email = default_member_email or ""
                if not auto_confirm:
                    if sys.stdin is None or not sys.stdin.isatty():
                        raise RuntimeError(
                            "Interactive input required for team creation. Use -y and provide --member-email.")
                    try:
                        tn = input(f"Team name [{default_team_name}]: ").strip()
                        if tn:
                            final_team_name = tn
                        af = input(f"Affiliation [{default_affiliation}]: ").strip()
                        if af:
                            final_affiliation = af
                        mn = input(f"First member name [{default_member_name}]: ").strip()
                        if mn:
                            final_member_name = mn
                        me = input(f"First member email [{default_member_email or ''}]: ").strip()
                        if me:
                            final_member_email = me
                    except EOFError:
                        pass
                if not final_member_email:
                    raise RuntimeError(
                        "Team creation cancelled: first member email is required!")

                try:
                    r = self.auth.register_team(
                        username=username,
                        display_name=display_name or username,
                        password=password,
                        team_name=final_team_name,
                        affiliation=final_affiliation,
                        first_member_name=final_member_name,
                        first_member_email=final_member_email,)
                    if self.debug:
                        dump_http(
                            r,
                            note=(
                                "/auth/register-team OK"
                                if 200 <= r.status_code < 300
                                else "/auth/register-team RESP"
                            ),)
                    r.raise_for_status()
                except requests.HTTPError as e:
                    status3 = (
                        e.response.status_code
                        if getattr(e, "response", None) is not None
                        else None )
                    if status3 == 400:
                        print(
                            f"Have you signed up with a .edu.au domain? ERR: {e}")
                        sys.exit(1)
                    raise
                info = self._persist_login(username, r)
                try:
                    team_info = self.competition.get(username) or {}
                    data = team_info.get("data") if isinstance(team_info, dict) else None
                    if isinstance(data, dict):
                        inv = data.get("invite_code")
                        if inv:
                            self._update_config_team_invite_code(inv)
                except Exception:
                    pass
                return info

            raise

    def _auth_headers(self, username: str) -> Dict[str, str]:
        tok = self.tokens.get(username)
        if not tok:
            raise RuntimeError(
                f"No token found for {username}. Run: twooter login {username}")
        token, token_type = tok
        if token_type and token_type.startswith("Cookie:"):
            cookie_name = token_type.split(":", 1)[1] or "session"
            return {"Cookie": f"{cookie_name}={token}"}
        return {"Authorization": f"Bearer {token}"}

    def create_post(
        self, username: str, content: str, parent_id: Optional[int] = None
    ) -> Dict[str, Any]:
        return self.twoots.create(username, content, parent_id=parent_id)

    def agent(self, username: str) -> Agent:
        return Agent(username, self.twoots, self.users, self.notifications)



from typing import Any, Dict, Optional


class CompetitionAPI:
    def __init__(self, api_session, headers_for_username):
        self._api = api_session
        self._headers_for = headers_for_username
        self._base = "/competition"
        # Team endpoints are under /competition/team
        self._team = f"{self._base}/team"

    # Public-ish (team admin role on user for get()):
    def get(self, username: Optional[str] = None) -> Dict[str, Any]:
        headers = self._headers_for(username) if username else None
        r = self._api.get(self._team, headers=headers)
        r.raise_for_status()
        return r.json()

    def verify_get(self, token: str) -> Dict[str, Any]:
        r = self._api.get(f"{self._base}/verify?token={token}")
        r.raise_for_status()
        return r.json()

    def verify_post(self, name: str, email: str, token: str,
                    consent: bool, is_student: bool, is_18: bool) -> Dict[str, Any]:
        body = {
            "name": name,
            "email": email,
            "token": token,
            "consents_to_experiment_according_to_ethics_document": consent,
            "is_australian_uni_student": is_student,
            "is_at_least_18_years_old": is_18,
        }
        r = self._api.post(f"{self._base}/verify", json_body=body)
        r.raise_for_status()
        return r.json()

    # Team-admin endpoints under /team
    def team_update(self, username: str, name: str, affiliation: str) -> Dict[str, Any]:
        r = self._api.post(f"{self._team}/", json_body={"name": name, "affiliation": affiliation}, headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def team_members(self, username: str) -> Dict[str, Any]:
        r = self._api.get(f"{self._team}/members", headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def team_member_create(self, username: str, name: str, email: str) -> Dict[str, Any]:
        r = self._api.post(f"{self._team}/members", json_body={"name": name, "email": email}, headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def team_member_get(self, username: str, member_id: int) -> Dict[str, Any]:
        r = self._api.get(f"{self._team}/members/{member_id}", headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def team_member_update(self, username: str, member_id: int, name: str, email: str) -> Dict[str, Any]:
        r = self._api.post(f"{self._team}/members/{member_id}", json_body={"name": name, "email": email}, headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def team_member_resend(self, username: str, member_id: int) -> Dict[str, Any]:
        r = self._api.post(f"{self._team}/members/{member_id}/resend-verification", json_body={}, headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def team_member_delete(self, username: str, member_id: int) -> Dict[str, Any]:
        r = self._api.session.delete(self._api.url(f"{self._team}/members/{member_id}"), headers=self._headers_for(username))
        if 200 <= r.status_code < 300:
            return r.json()
        r.raise_for_status()
        return r.json()

    def users(self, username: str, q: Optional[str] = None, admins: Optional[bool] = None) -> Dict[str, Any]:
        params = []
        if q:
            params.append(f"q={q}")
        if admins is not None:
            params.append(f"admins={'true' if admins else 'false'}")
        qs = ("?" + "&".join(params)) if params else ""
        r = self._api.get(f"{self._team}/users{qs}", headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def promote(self, username: str, target_username: str) -> Dict[str, Any]:
        r = self._api.post(f"{self._team}/promote/{target_username}", json_body={}, headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def demote(self, username: str, target_username: str) -> Dict[str, Any]:
        r = self._api.post(f"{self._team}/demote/{target_username}", json_body={}, headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def rotate_invite_code(self, username: str) -> Dict[str, Any]:
        r = self._api.post(f"{self._team}/rotate-invite-code", json_body={}, headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()



from typing import Any, Dict, Optional


class Agent:
    def __init__(self, username: str, twoots_api, users_api, notifications_api):
        self.username = username
        self._twoots = twoots_api
        self._users = users_api
        self._notifications = notifications_api

    # Posts
    def post(self, content: str, parent_id: Optional[int] = None, embed: Optional[str] = None, media: Optional[list[str]] = None) -> Dict[str, Any]:
        return self._twoots.create(self.username, content, parent_id=parent_id, embed=embed, media=media)

    def like(self, post_id: int) -> Dict[str, Any]:
        return self._twoots.like(self.username, post_id)

    def unlike(self, post_id: int) -> Dict[str, Any]:
        return self._twoots.unlike(self.username, post_id)

    def repost(self, post_id: int) -> Dict[str, Any]:
        return self._twoots.repost(self.username, post_id)

    def unrepost(self, post_id: int) -> Dict[str, Any]:
        return self._twoots.unrepost(self.username, post_id)

    def delete_post(self, post_id: int) -> Dict[str, Any]:
        return self._twoots.delete(self.username, post_id)

    # Social
    def follow(self, target_username: str) -> Dict[str, Any]:
        return self._users.follow(self.username, target_username)

    def unfollow(self, target_username: str) -> Dict[str, Any]:
        return self._users.unfollow(self.username, target_username)

    # Notifications
    def notifications(self) -> Dict[str, Any]:
        return self._notifications.list(self.username)

    def notifications_unread(self) -> Dict[str, Any]:
        return self._notifications.unread(self.username)

    def notifications_count(self) -> Dict[str, Any]:
        return self._notifications.count(self.username)

    def notifications_count_unread(self) -> Dict[str, Any]:
        return self._notifications.count_unread(self.username)

    def mark_read(self, notification_id: int) -> Dict[str, Any]:
        return self._notifications.mark_read(self.username, notification_id)

    def mark_unread(self, notification_id: int) -> Dict[str, Any]:
        return self._notifications.mark_unread(self.username, notification_id)

    def delete_notification(self, notification_id: int) -> Dict[str, Any]:
        return self._notifications.delete(self.username, notification_id)

    def clear_notifications(self) -> Dict[str, Any]:
        return self._notifications.clear(self.username)




from __future__ import annotations

# Depending on ur config, VSCode might indicate this is not being used. It is.
from .runner import main
from .client import TwooterClient
from .config import Config



from __future__ import annotations

from typing import Any, Dict, Optional, Tuple

import requests

from .util import dump_http


def extract_token(resp: requests.Response) -> Tuple[str, Optional[str], Optional[int]]:
    token = None
    token_type = None
    expires_at = None

    data = {}
    try:
        data = resp.json()
    except Exception:
        pass

    tok = data.get("access_token") or data.get("token")
    if tok:
        token = tok
        token_type = data.get("token_type") or "Bearer"
        expires_in = data.get("expires_in")
        if isinstance(expires_in, (int, float)):
            import time
            expires_at = int(time.time() + int(expires_in))

    if not token:
        preferred = ("session", "access_token", "token")
        chosen_name = None
        for name in preferred:
            if name in resp.cookies:
                chosen_name = name
                token = resp.cookies.get(name)
                break
        if not token:
            for c in resp.cookies:
                chosen_name = c.name
                token = c.value
                break
        if token:
            token_type = f"Cookie:{chosen_name or 'session'}"

    if not token:
        raise ValueError("Login returned no recognisable token or session cookie.")
    return token, token_type, expires_at


class ApiSession:
    def __init__(self, base_url: str, debug: bool = False) -> None:
        self.base_url = base_url.rstrip("/")
        self.session = requests.Session()
        self.debug = debug

    def url(self, path: str) -> str:
        return f"{self.base_url}{path}"

    # HTTP POST, not making an actual post on the website :P
    def post(self, path: str, *, json_body: Dict[str, Any], headers: Optional[Dict[str, str]] = None) -> requests.Response:
        r = self.session.post(self.url(path), json=json_body, headers=headers)
        if self.debug:
            dump_http(r, note=f"{path} {'OK' if 200 <= r.status_code < 300 else 'RESP'}")
        return r

    # HTTP GET, idk how you could get confused with this one lol
    def get(self, path: str, *, headers: Optional[Dict[str, str]] = None) -> requests.Response:
        r = self.session.get(self.url(path), headers=headers)
        if self.debug:
            dump_http(r, note=f"{path} {'OK' if 200 <= r.status_code < 300 else 'RESP'}")
        return r



from typing import Any, Dict


class NotificationsAPI:
    def __init__(self, api_session, headers_for_username):
        self._api = api_session
        self._headers_for = headers_for_username

    def list(self, username: str) -> Dict[str, Any]:
        r = self._api.get("/notifications", headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def unread(self, username: str) -> Dict[str, Any]:
        r = self._api.get("/notifications/unread", headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def count(self, username: str) -> Dict[str, Any]:
        r = self._api.get("/notifications/count", headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def count_unread(self, username: str) -> Dict[str, Any]:
        r = self._api.get("/notifications/unread/count", headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def mark_read(self, username: str, notification_id: int) -> Dict[str, Any]:
        r = self._api.post(f"/notifications/{notification_id}/mark-read", json_body={}, headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def mark_unread(self, username: str, notification_id: int) -> Dict[str, Any]:
        r = self._api.post(f"/notifications/{notification_id}/mark-unread", json_body={}, headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def delete(self, username: str, notification_id: int) -> Dict[str, Any]:
        r = self._api.session.delete(self._api.url(f"/notifications/{notification_id}"), headers=self._headers_for(username))
        if 200 <= r.status_code < 300:
            return r.json()
        r.raise_for_status()
        return r.json()

    def clear(self, username: str) -> Dict[str, Any]:
        r = self._api.session.delete(self._api.url("/notifications/clear"), headers=self._headers_for(username))
        if 200 <= r.status_code < 300:
            return r.json()
        r.raise_for_status()
        return r.json()




from typing import Any, Dict, Union


class UsersAPI:
    def __init__(self, api_session, headers_for_username):
        self._api = api_session
        self._headers_for = headers_for_username
        self._base = "/users"

    def _fmt_ident(self, ident: Union[str, int]) -> str:
        if isinstance(ident, int):
            return str(ident)
        s = str(ident).strip()
        # Treat purely numeric strings as IDs
        if s.isdigit():
            return s
        # Ensure '@' prefix for usernames, because that's what we decided on lol
        return s if s.startswith("@") else f"@{s}"

    def _extract_id(self, payload: Dict[str, Any]) -> int:
        """Try to extract a numeric user id from a GET /users/@name response."""
        if isinstance(payload, dict):
            data = payload.get("data") if isinstance(payload.get("data"), dict) else payload
            if isinstance(data, dict) and "id" in data:
                try:
                    return int(data["id"])
                except Exception:
                    pass
        raise ValueError("Unable to resolve user id from response payload")

    def me(self, username: str) -> Dict[str, Any]:
        r = self._api.get(f"{self._base}/me", headers=self._headers_for(username))
        r.raise_for_status()
        return r.json()

    def update_me(self, username: str, display_name: str, bio: str) -> Dict[str, Any]:
        r = self._api.post(
            f"{self._base}/me",
            json_body={"display_name": display_name, "bio": bio},
            headers=self._headers_for(username),
        )
        r.raise_for_status()
        return r.json()

    def get(self, username_or_id: Union[str, int]) -> Dict[str, Any]:
        ident = self._fmt_ident(username_or_id)
        r = self._api.get(f"{self._base}/{ident}/")
        r.raise_for_status()
        return r.json()

    def activity(self, username_or_id: Union[str, int]) -> Dict[str, Any]:
        ident = self._fmt_ident(username_or_id)
        r = self._api.get(f"{self._base}/{ident}/activity")
        r.raise_for_status()
        return r.json()

    def follows(self, username_or_id: Union[str, int]) -> Dict[str, Any]:
        ident = self._fmt_ident(username_or_id)
        r = self._api.get(f"{self._base}/{ident}/follows")
        r.raise_for_status()
        return r.json()

    def followers(self, username_or_id: Union[str, int]) -> Dict[str, Any]:
        ident = self._fmt_ident(username_or_id)
        r = self._api.get(f"{self._base}/{ident}/followers")
        r.raise_for_status()
        return r.json()

    def follow(self, agent_username: str, target_username_or_id: Union[str, int]) -> Dict[str, Any]:
        target_path = self._fmt_ident(target_username_or_id)
        r = self._api.post(
            f"{self._base}/{target_path}/follow",
            json_body={},
            headers=self._headers_for(agent_username),
        )
        r.raise_for_status()
        return r.json()

    def unfollow(self, agent_username: str, target_username_or_id: Union[str, int]) -> Dict[str, Any]:
        target_path = self._fmt_ident(target_username_or_id)
        r = self._api.session.delete(
            self._api.url(f"{self._base}/{target_path}/follow"),
            headers=self._headers_for(agent_username),
        )
        if 200 <= r.status_code < 300:
            return r.json()
        r.raise_for_status()
        return r.json()



from __future__ import annotations

import argparse
from datetime import datetime


FEED_KEYS = ["trending", "latest", "home", "explore"]


def parse_args() -> argparse.Namespace:
    p = argparse.ArgumentParser(
        description="Twooter CLI",
        epilog=(
            "Notes:\n"
            "- Use @username for user lookups; numeric IDs also accepted.\n"
            "- feeds -n limits results client-side.\n"
            f"- feeds keys: {', '.join(FEED_KEYS)}"
            "\n- auth whoami shows acting user's role and team info."
            "\n- auth token-info shows saved token type and expiry."
            "\n- On login failure, it notes the user may not exist, prompts to try the bot key (if set), then tries team_invite_code,"
            " and finally creates a team if needed. After creating a team, XDG config is updated with team_invite_code."
            "\n  Use -y and optionally --team-name/--affiliation/--member-name/--member-email for non-interactive flows."
        ),)
    p.add_argument("--config")
    p.add_argument("--debug", action="store_true")

    sub = p.add_subparsers(dest="command")

    def add_agent(sp):
        mg = sp.add_mutually_exclusive_group(required=False)
        mg.add_argument("--as", dest="as_user")
        mg.add_argument("--asindex", dest="as_index", type=int)

    login = sub.add_parser("login")
    lg = login.add_mutually_exclusive_group(required=False)
    lg.add_argument("--user")
    lg.add_argument("--index", type=int)
    login.add_argument("who", nargs="?")
    login.add_argument("-y", "--yes", action="store_true")
    login.add_argument("--team-name", dest="team_name")
    login.add_argument("--affiliation", dest="affiliation")
    login.add_argument("--member-name", dest="member_name")
    login.add_argument("--member-email", dest="member_email")

    users = sub.add_parser("users")
    users_sub = users.add_subparsers(dest="users_cmd", required=True)
    ug = users_sub.add_parser("get")
    ug.add_argument("identifier")
    ume = users_sub.add_parser("me")
    add_agent(ume)
    uup = users_sub.add_parser("update")
    add_agent(uup)
    uup.add_argument("--display-name", required=True)
    uup.add_argument("--bio", required=True)
    ua = users_sub.add_parser("activity")
    ua.add_argument("identifier")
    ufo = users_sub.add_parser("follows")
    ufo.add_argument("identifier")
    ufr = users_sub.add_parser("followers")
    ufr.add_argument("identifier")
    uf = users_sub.add_parser("follow")
    add_agent(uf)
    uf.add_argument("target")
    uu = users_sub.add_parser("unfollow")
    add_agent(uu)
    uu.add_argument("target")

    tw = sub.add_parser("twoots")
    tw_sub = tw.add_subparsers(dest="twoots_cmd", required=True)
    twc = tw_sub.add_parser("create")
    add_agent(twc)
    twc.add_argument("--content", required=True)
    twc.add_argument("--parent-id", type=int)
    twc.add_argument("--embed")
    twc.add_argument("--media", nargs="*")
    twg = tw_sub.add_parser("get")
    twg.add_argument("post_id", type=int)
    twr = tw_sub.add_parser("replies")
    twr.add_argument("post_id", type=int)
    twl = tw_sub.add_parser("like")
    add_agent(twl)
    twl.add_argument("post_id", type=int)
    twul = tw_sub.add_parser("unlike")
    add_agent(twul)
    twul.add_argument("post_id", type=int)
    twrp = tw_sub.add_parser("repost")
    add_agent(twrp)
    twrp.add_argument("post_id", type=int)
    twurp = tw_sub.add_parser("unrepost")
    add_agent(twurp)
    twurp.add_argument("post_id", type=int)
    twd = tw_sub.add_parser("delete")
    add_agent(twd)
    twd.add_argument("post_id", type=int)

    twe = tw_sub.add_parser("embed")
    twe.add_argument("post_id", type=int)

    twald = tw_sub.add_parser("allowed-link-domains")

    twrep = tw_sub.add_parser("report")
    add_agent(twrep)
    twrep.add_argument("post_id", type=int)
    twrep.add_argument("--reason", required=True)

    twvis = tw_sub.add_parser("visibility")
    add_agent(twvis)
    twvis.add_argument("post_id", type=int)
    twvis.add_argument("--visibility", required=True, choices=["default", "safe", "manually_hidden", "needs-review"])

    twpi = tw_sub.add_parser("prompt-injection")
    add_agent(twpi)
    twpi.add_argument("post_id", type=int)
    twpi.add_argument("--value", required=True, choices=["true", "false"], help="Enable or disable prompt injection flag")

    no = sub.add_parser("notifications")
    no_sub = no.add_subparsers(dest="notifications_cmd", required=True)
    nli = no_sub.add_parser("list")
    add_agent(nli)
    nur = no_sub.add_parser("unread")
    add_agent(nur)
    nco = no_sub.add_parser("count")
    add_agent(nco)
    ncu = no_sub.add_parser("count-unread")
    add_agent(ncu)
    nmr = no_sub.add_parser("mark-read")
    add_agent(nmr)
    nmr.add_argument("id", type=int)
    nmu = no_sub.add_parser("mark-unread")
    add_agent(nmu)
    nmu.add_argument("id", type=int)
    ndl = no_sub.add_parser("delete")
    add_agent(ndl)
    ndl.add_argument("id", type=int)
    ncl = no_sub.add_parser("clear")
    add_agent(ncl)

    tg = sub.add_parser("tags")
    tg_sub = tg.add_subparsers(dest="tags_cmd", required=True)
    tg_sub.add_parser("trending")
    tgl = tg_sub.add_parser("lookup")
    tgl.add_argument("name")
    tga = tg_sub.add_parser("latest")
    tga.add_argument("name")

    se = sub.add_parser("search")
    se.add_argument("query")

    _feed_keys_str = ", ".join(FEED_KEYS)
    fe = sub.add_parser(
        "feeds",
        description=(
            f"Available feed keys: {_feed_keys_str}. "
            "'home' and 'explore' require authentication (use --as/--asindex)."
        ),)
    add_agent(fe)
    fe.add_argument("key", nargs="?")
    fe.add_argument("--at")
    fe.add_argument("-n", "--n", type=int)
    fe.add_argument("--list", action="store_true", help="List available feeds instead of fetching one")

    co = sub.add_parser("competition")
    co_sub = co.add_subparsers(dest="competition_cmd", required=True)
    cot = co_sub.add_parser("team")
    add_agent(cot)
    cou = co_sub.add_parser("team-update")
    add_agent(cou)
    cou.add_argument("--name", required=True)
    cou.add_argument("--affiliation", required=True)
    com = co_sub.add_parser("members")
    add_agent(com)
    comc = co_sub.add_parser("member-create")
    add_agent(comc)
    comc.add_argument("--name", required=True)
    comc.add_argument("--email", required=True)
    comg = co_sub.add_parser("member-get")
    add_agent(comg)
    comg.add_argument("member_id", type=int)
    comu = co_sub.add_parser("member-update")
    add_agent(comu)
    comu.add_argument("member_id", type=int)
    comu.add_argument("--name", required=True)
    comu.add_argument("--email", required=True)
    comr = co_sub.add_parser("member-resend")
    add_agent(comr)
    comr.add_argument("member_id", type=int)
    comd = co_sub.add_parser("member-delete")
    add_agent(comd)
    comd.add_argument("member_id", type=int)
    couu = co_sub.add_parser("users")
    add_agent(couu)
    couu.add_argument("--q")
    couu.add_argument("--admins", choices=["true", "false"])
    cop = co_sub.add_parser("promote")
    add_agent(cop)
    cop.add_argument("target")
    cod = co_sub.add_parser("demote")
    add_agent(cod)
    cod.add_argument("target")
    cor = co_sub.add_parser("rotate-invite-code")
    add_agent(cor)
    covg = co_sub.add_parser("verify-get")
    covg.add_argument("token")
    covp = co_sub.add_parser("verify-post")
    covp.add_argument("--name", required=True)
    covp.add_argument("--email", required=True)
    covp.add_argument("--token", required=True)
    covp.add_argument("--consent", action="store_true")
    covp.add_argument("--student", action="store_true")
    covp.add_argument("--age18", action="store_true")

    au = sub.add_parser(
        "auth",
        description=(
            "Authentication and account operations.\n"
            "Subcommands: change-password, logout, register-team (create TEAM_ADMIN), whoami (show role/team), token-info (saved token deets)."
        ),)
    au_sub = au.add_subparsers(dest="auth_cmd", required=True)
    auc = au_sub.add_parser("change-password")
    add_agent(auc)

    # epic new personas DB management command/sub
    pe = sub.add_parser(
        "personas",
        description=(
            "Manage local personas.db entries (no network)."
        ),)
    pe_sub = pe.add_subparsers(dest="personas_cmd", required=True)
    pea = pe_sub.add_parser("add")
    pea.add_argument("--username", required=True)
    pea.add_argument("--password", required=True)
    pea.add_argument("--email", required=True)
    pea.add_argument("--display-name")
    pea.add_argument("--team-invite-code")
    pel = pe_sub.add_parser("list")
    ped = pe_sub.add_parser("delete")
    ped.add_argument("identifier", help="username or email")
    auc.add_argument("--new-password", required=True)
    aul = au_sub.add_parser("logout")
    add_agent(aul)

    aur = au_sub.add_parser("register-team")
    rg = aur.add_mutually_exclusive_group(required=False)
    rg.add_argument("--user")
    rg.add_argument("--index", type=int)
    aur.add_argument("who", nargs="?")
    aur.add_argument("--team-name", required=True)
    aur.add_argument("--affiliation", required=True)
    aur.add_argument("--member-name")
    aur.add_argument("--member-email")

    auw = au_sub.add_parser("whoami")
    add_agent(auw)
    aut = au_sub.add_parser("token-info")
    add_agent(aut)

    return p.parse_args()
